package sshwrapper

import (
	"context"
	"errors"
	"fmt"
	"log"
	"os/exec"
	"strings"
	"sync"
	"sync/atomic"
	"syscall"
	"time"

	"github.com/it-novum/nsta/nstautil"
)

const ProcessStopped = 0
const ProcessStarting = -1

// Conn manages the ssh connection with restart logic and dynamic port forwarding
// we try to avoid mutex here to prevent deadlocks or race conditions
type Conn struct {
	ConnOpts   ConnOpts
	ClientName string

	wg       sync.WaitGroup
	shutdown chan struct{}

	// if > 0 it holds the PID of the ssh process, otherwise a Process* const state
	pid     int64
	sshPort chan int

	cmd    *exec.Cmd
	output nstautil.SyncBuffer
}

// formatSSHCommand returns the full ssh command for exec
func (c *Conn) formatSSHCommand(localPort int) []string {
	args := c.ConnOpts.CommandBase()
	return append(args[:len(args)-1], []string{
		"-L",
		fmt.Sprintf("%d:127.0.0.1:%d", localPort, c.ConnOpts.RemotePort),
		"-N",
		args[len(args)-1],
	}...)
}

// sshProcess must be called by startSSHProcess
func (c *Conn) sshProcess(parent context.Context) {
	defer c.wg.Done()
	// we can safely set this to ProcessStopped, because the c.cmd.Wait func returns only if the process is dead
	// as syscall in c.cmd.Wait ensures this
	defer atomic.StoreInt64(&c.pid, ProcessStopped)

	ctx, cancel := context.WithCancel(parent)
	defer cancel()

	defer func() {
		c.cmd = nil
		// we keep c.output for logging
	}()

	port, err := nstautil.DynamicPort()
	if err != nil {
		log.Println(&Error{Name: c.ClientName, Message: "could not get a dynamic port for port forwarding", Err: err})
		return
	}
	cmdArgs := c.formatSSHCommand(port)

	log.Println(&Error{Name: c.ClientName, Message: fmt.Sprint(cmdArgs)})
	cmd := exec.CommandContext(ctx, cmdArgs[0], cmdArgs[1:]...)
	cmd.Stdin = strings.NewReader("")

	c.output.Reset()
	cmd.Stdout = &c.output
	cmd.Stderr = &c.output

	if err := cmd.Start(); err != nil {
		log.Println(&Error{Name: c.ClientName, Message: "could not start process", Err: err})
		return
	}

	c.cmd = cmd
	atomic.StoreInt64(&c.pid, int64(cmd.Process.Pid))
	c.wg.Add(1)
	// Test if the port is really open, then send the new port upstream
	go func() {
		defer c.wg.Done()

		tick := time.NewTicker(time.Millisecond * 200)
		defer tick.Stop()

		for {
			select {
			case <-ctx.Done():
				return
			case <-tick.C:
				if nstautil.ConnectionTest("127.0.0.1", port) {
					// we have to test if the context is canceled!
					// for example if NSTA is reloaded and nothing checks the Port channel for new data and
					// the ssh process should be stopped this would hang endless or send the port information
					// when already another process has been started
					select {
					case <-ctx.Done():
						return

					// Send dynamic port to satellite instance
					case c.sshPort <- port:
						return
					}
				}
			}
		}
	}()

	if err := c.cmd.Wait(); err != nil {
		log.Println(&Error{Name: c.ClientName, Message: "SSH process stopped", Err: err})
	}
}

func (c *Conn) startSSHProcess(ctx context.Context) {
	if atomic.LoadInt64(&c.pid) > 0 || c.cmd != nil {
		panic("!!!BUG: invalid ssh process state!!!")
	}

	// we have to set this here so we don't have problems with race conditions
	// checkProcess must need to know that the process is starting before it is actually started
	c.wg.Add(1)
	atomic.StoreInt64(&c.pid, ProcessStarting)

	// to be sure that upstream knows that we are not connected
	// and also to ensure that some upstream is really listening before we beginning
	// we send port number 0 which is signal that no ssh connection is established
	// a successful ssh connection and port forwarding doesn't mean that a tcp connection to the real endpoint
	// is actually possible, because the destination service on the remote machine can be stopped
	// also we have to check if the context has been canceled if upstream decides to suddenly stop or so
	select {
	case <-ctx.Done():
		// DO NOT USE DEFER!
		atomic.StoreInt64(&c.pid, ProcessStopped)
		c.wg.Done()
		return
	case _, more := <-c.shutdown:
		if !more {
			return
		}
		// DO NOT USE DEFER!
		atomic.StoreInt64(&c.pid, ProcessStopped)
		c.wg.Done()
		return
	case c.sshPort <- 0:
		go c.sshProcess(ctx)
	}
}

// checkProcess and start it if it is not starting or already running
func (c *Conn) checkProcess(ctx context.Context) {
	if atomic.LoadInt64(&c.pid) == ProcessStopped {
		// c.output is only set if the process started successfully
		if c.output.Len() > 0 {
			log.Println(&Error{Name: c.ClientName, Message: fmt.Sprintf("stopped unexpectedly with output: %s", c.output.String())})
		}
		c.startSSHProcess(ctx)
	}
}

func (c *Conn) run(parent context.Context) {
	defer c.wg.Done()

	ctx, cancel := context.WithCancel(parent)
	defer cancel()

	defer func() {
		log.Println(&Error{Name: c.ClientName, Message: "stop ssh process"})

		// we use c.pid instead of c.cmd.Process.Kill because it unsafe!!
		// it could be nil or in some other unpredictable state that is influenced by another thread
		// therefore we use the thread safe c.pid (using atomic)

		tryCount := 0
		for pid := atomic.LoadInt64(&c.pid); pid != ProcessStopped; {

			if pid == ProcessStarting {
				// pid -1 could happen if the SSH process was not spawned yet.
				// Das hier passiert NICHT beim Stoppen des SSH-Prozesses! Wenn der SSH Prozess nicht richtig stoppen würde, würde der NSTA nicht in dieser Schleife hängen
				// Der Start des SSH Prozess ist es, der Probleme macht. Der SSH Prozess wurde gestartet, aber wurde noch nicht als läuft marktiert. Hier wird darauf gewartet
				// das der SSH Prozess auch wirklich läuft.

				log.Println(&Error{Name: c.ClientName, Message: "wait for startup of ssh to finish before we can stop it"})
				time.Sleep(time.Second)
				continue
			}

			tryCount++
			if tryCount == 1 {
				if err := syscall.Kill(int(pid), syscall.SIGTERM); errors.Is(err, syscall.ESRCH) {
					// already dead, probably a timing issue, the process could stopped after we fetched the pid
					break
				}
			} else if tryCount > 100 {
				// after 10 seconds we try to kill it
				if err := syscall.Kill(int(pid), syscall.SIGKILL); errors.Is(err, syscall.ESRCH) {
					// already dead, probably a timing issue, the process could stopped after we fetched the pid
					break
				}
			}
			time.Sleep(time.Millisecond * 100)
		}
	}()

	timer := time.NewTicker(c.ConnOpts.Reconnect)
	defer timer.Stop()

	resetBuffer := time.NewTimer(time.Hour * 4)
	defer resetBuffer.Stop()

	c.checkProcess(ctx)

	for {
		select {
		case <-ctx.Done():
			return
		case _, more := <-c.shutdown:
			if !more {
				return
			}
		case <-timer.C:
			c.checkProcess(ctx)
		case <-resetBuffer.C:
			// Reset ssh output buffer after 4 hours to avoid memory leak
			c.output.Reset()
		}
	}
}

// Start the background worker, returns the channel that receives port number updates
// the process will not start if the port is not fetched and will always return 0 for the first time
// it will not be closed after Stop or context cancel
func (c *Conn) Start(ctx context.Context) (<-chan int, error) {
	if verify := c.ConnOpts.Verify(); verify != "" {
		return nil, &Error{Name: c.ClientName, Message: verify}
	}

	if c.ConnOpts.Reconnect == 0 {
		c.ConnOpts.Reconnect = time.Second * 5
	}

	c.shutdown = make(chan struct{})
	c.pid = ProcessStopped

	// Channel used to transfer the dynamic port to the satellite instance
	c.sshPort = make(chan int)

	c.wg.Add(1)
	go c.run(ctx)

	return c.sshPort, nil
}

// Stop the ssh process
func (c *Conn) Stop() {
	close(c.shutdown)
	c.wg.Wait()
}
